"""drivers/sqlite.py

SQLite-backed raster driver. Metadata is stored in an SQLite database, raster data is assumed
to be present on disk.
"""

from typing import Any, Iterator
import os
import tempfile
import shutil
import operator
import logging
import contextlib
import urllib.parse as urlparse

from cachetools import cachedmethod, TTLCache

from terracotta import get_settings, exceptions
from terracotta.drivers.sqlite import SQLiteDriver
from terracotta.profile import trace

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def convert_exceptions(msg: str) -> Iterator:
    """Convert internal sqlite and boto exceptions to our InvalidDatabaseError"""
    import sqlite3
    import botocore.exceptions
    try:
        yield
    except (sqlite3.OperationalError, botocore.exceptions.ClientError) as exc:
        raise exceptions.InvalidDatabaseError(msg) from exc


def _update_from_s3(remote_path: str, local_path: str) -> None:
    import boto3

    parsed_remote_path = urlparse.urlparse(remote_path)
    bucket_name, key = parsed_remote_path.netloc, parsed_remote_path.path.strip('/')

    if parsed_remote_path.scheme != 's3':
        raise ValueError('Expected s3:// URL')

    s3 = boto3.resource('s3')
    obj = s3.Object(bucket_name, key)
    obj_bytes = obj.get()['Body']

    # copy over existing database; this is somewhat safe since it is read-only
    # NOTE: replace with Connection.backup after switching to Python 3.7
    with open(local_path, 'wb') as f:
        shutil.copyfileobj(obj_bytes, f)


class RemoteSQLiteDriver(SQLiteDriver):
    """An SQLite-backed raster driver, where the database file is stored remotely on S3.

    Assumes raster data to be present in separate GDAL-readable files on disk or remotely.
    Stores metadata and paths to raster files in SQLite.

    See also:

        :class:`~terracotta.drivers.sqlite.SQLiteDriver` for the local version of this
        driver.

    The SQLite database is simply a file that can be stored together with the actual
    raster files on S3. Before handling the first request, this driver will download a
    temporary copy of the remote database file. It is thus not feasible for large databases.

    The local database copy will be updated in regular intervals defined by
    :attr:`~terracotta.config.TerracottaSettings.REMOTE_DB_CACHE_TTL`.

    Warning:

        This driver is read-only. Any attempts to use the create, insert, or delete methods
        will throw a NotImplementedError.

    """
    path: str

    def __init__(self, remote_path: str) -> None:
        """Initialize the RemoteSQLiteDriver.

        This should not be called directly, use :func:`~terracotta.get_driver` instead.

        Arguments:

            remote_path: S3 URL in the form ``s3://bucket/key`` to remote SQLite database
                (has to exist).

        """
        settings = get_settings()

        self.__rm = os.remove  # keep reference to use in __del__

        os.makedirs(settings.REMOTE_DB_CACHE_DIR, exist_ok=True)
        local_db_file = tempfile.NamedTemporaryFile(
            dir=settings.REMOTE_DB_CACHE_DIR,
            prefix='tc_s3_db_',
            suffix='.sqlite',
            delete=False
        )
        local_db_file.close()

        self._remote_path: str = str(remote_path)
        self._checkdb_cache = TTLCache(maxsize=1, ttl=settings.REMOTE_DB_CACHE_TTL)

        super().__init__(local_db_file.name)

    @cachedmethod(operator.attrgetter('_checkdb_cache'))
    @convert_exceptions('Could not retrieve database from S3')
    @trace('download_db_from_s3')
    def _update_db(self, remote_path: str, local_path: str) -> None:
        logger.debug('Remote database cache expired, re-downloading')
        _update_from_s3(remote_path, local_path)

    def _connection_callback(self) -> None:
        self._update_db(self._remote_path, self.path)
        super()._connection_callback()

    def create(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError('Remote SQLite databases are read-only')

    def insert(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError('Remote SQLite databases are read-only')

    def delete(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError('Remote SQLite databases are read-only')

    def __del__(self) -> None:
        """Clean up temporary database upon exit"""
        self.__rm(self.path)



"""handlers/keys.py

Handle /keys API endpoint.
"""

from typing import List, Dict

from terracotta import get_settings, get_driver
from terracotta.profile import trace


@trace('keys_handler')
def keys() -> List[Dict[str, str]]:
    """List available keys, in order"""
    settings = get_settings()
    driver = get_driver(settings.DRIVER_PATH, provider=settings.DRIVER_PROVIDER)

    response = []
    for key, description in driver.get_keys().items():
        response_row = {'key': key}
        if description:
            response_row['description'] = description

        response.append(response_row)

    return response



"""handlers/metadata.py

Handle /metadata API endpoint.
"""

from typing import Mapping, Sequence, Dict, Any, Union
from collections import OrderedDict

from terracotta import get_settings, get_driver
from terracotta.profile import trace


@trace('metadata_handler')
def metadata(keys: Union[Sequence[str], Mapping[str, str]]) -> Dict[str, Any]:
    """Returns all metadata for a single dataset"""
    settings = get_settings()
    driver = get_driver(settings.DRIVER_PATH, provider=settings.DRIVER_PROVIDER)
    metadata = driver.get_metadata(keys)
    metadata['keys'] = OrderedDict(zip(driver.key_names, keys))
    return metadata



"""server/app.py

Instantiated version of the Flask API.
"""

from terracotta import get_settings, logs
from terracotta.server import create_app

settings = get_settings()

logs.set_logger(
    settings.LOGLEVEL,
    catch_warnings=True
)

app = create_app(
    debug=settings.DEBUG,
    profile=settings.FLASK_PROFILE
)



"""profile.py

Decorators for performance tracing.
"""

from typing import Iterator

import traceback
import contextlib

from terracotta import get_settings


@contextlib.contextmanager
def trace(description: str) -> Iterator:
    settings = get_settings()
    if settings.XRAY_PROFILE:
        from aws_xray_sdk.core import xray_recorder
        try:
            subsegment = xray_recorder.begin_subsegment(description)
            yield subsegment
        except Exception as exc:
            stack = traceback.extract_stack()
            subsegment.add_exception(exc, stack)
            raise
        finally:
            xray_recorder.end_subsegment()
    else:
        yield



"""scripts/serve.py

Use Flask development server to serve up raster files or database locally.
"""

from typing import Any, Tuple, Sequence
import os
import tempfile
import logging

import click
import tqdm

from terracotta.scripts.click_types import RasterPattern, RasterPatternType
from terracotta.scripts.http_utils import find_open_port

logger = logging.getLogger(__name__)


@click.command('serve', short_help='Serve rasters through a local Flask development server.')
@click.option('-d', '--database', required=False, default=None, help='Database to serve from.')
@click.option('-r', '--raster-pattern', type=RasterPattern(), required=False, default=None,
              help='A format pattern defining paths and keys of the raster files to serve.')
@click.option('--rgb-key', default=None,
              help='Key to use for RGB compositing [default: last key in pattern]. '
                   'Has no effect if -r/--raster-pattern is not given.')
@click.option('--debug', is_flag=True, default=False, help='Enable Flask debugging.')
@click.option('--profile', is_flag=True, default=False, help='Enable Flask profiling.')
@click.option('--database-provider', default=None,
              help='Specify the driver to use to read database [default: auto detect].')
@click.option('--allow-all-ips', is_flag=True, default=False,
              help='Allow connections from outside IP addresses. Use with care!')
@click.option('--port', type=click.INT, default=None,
              help='Port to use [default: first free port between 5000 and 5099].')
def serve(database: str = None,
          raster_pattern: RasterPatternType = None,
          debug: bool = False,
          profile: bool = False,
          database_provider: str = None,
          allow_all_ips: bool = False,
          port: int = None,
          rgb_key: str = None) -> None:
    """Serve rasters through a local Flask development server.

    Either -d/--database or -r/--raster-pattern must be given.

    Example:

        $ terracotta serve -r /path/to/rasters/{name}/{date}_{band}_{}.tif

    The empty group {} is replaced by a wildcard matching anything (similar to * in glob patterns).

    This command is a data exploration tool and not meant for production use. Deploy Terracotta as
    a WSGI or serverless app instead.
    """
    from terracotta import get_driver, update_settings
    from terracotta.server import create_app

    if (database is None) == (raster_pattern is None):
        raise click.UsageError('Either --database or --raster-pattern must be given')

    if raster_pattern is not None:
        dbfile = tempfile.NamedTemporaryFile(suffix='.sqlite', delete=False)
        dbfile.close()

        keys, raster_files = raster_pattern

        if rgb_key is not None:
            if rgb_key not in keys:
                raise click.BadParameter('RGB key not found in raster pattern')

            # re-order keys
            rgb_idx = keys.index(rgb_key)

            def push_to_last(seq: Sequence[Any], index: int) -> Tuple[Any, ...]:
                return (*seq[:index], *seq[index + 1:], seq[index])

            keys = list(push_to_last(keys, rgb_idx))
            raster_files = {push_to_last(k, rgb_idx): v for k, v in raster_files.items()}

        driver = get_driver(dbfile.name, provider='sqlite')
        driver.create(keys)

        with driver.connect():
            click.echo('')
            for key, filepath in tqdm.tqdm(raster_files.items(), desc="Ingesting raster files"):
                driver.insert(key, filepath, skip_metadata=True)
            click.echo('')

        database = dbfile.name

    update_settings(DRIVER_PATH=database, DRIVER_PROVIDER=database_provider,
                    DEBUG=debug, FLASK_PROFILE=profile)

    # find suitable port
    port_range = [port] if port is not None else range(5000, 5100)
    port = find_open_port(port_range)
    if port is None:
        click.echo(f'Could not find open port to bind to (ports tried: {port_range})', err=True)
        raise click.Abort()

    host = '0.0.0.0' if allow_all_ips else 'localhost'

    server_app = create_app(debug=debug, profile=profile)

    if os.environ.get('TC_TESTING'):
        return

    server_app.run(port=port, host=host, threaded=False)  # pragma: no cover



"""logs.py

Provides logging utilities.
"""

from typing import Any
import logging
import logging.handlers

try:
    import colorlog
    use_colors = True
except ImportError:  # pragma: no cover
    use_colors = False


LEVEL_PREFIX = {
    'DEBUG': 'd',
    'INFO': '+',
    'WARNING': '!',
    'ERROR': '-',
    'CRITICAL': 'x'
}

LOG_COLORS = {
    'DEBUG': 'white,bold',
    'INFO': 'green,bold',
    'WARNING': 'yellow,bold',
    'ERROR': 'red,bold',
    'CRITICAL': 'red,bold',
}


def set_logger(level: str, catch_warnings: bool = False) -> logging.Logger:
    """Initialize loggers"""
    level = level.upper()

    package_logger = logging.getLogger('terracotta')
    package_logger.setLevel(level)

    # stream handler
    ch = logging.StreamHandler()
    ch_fmt: logging.Formatter

    if use_colors:
        fmt = ' {log_color!s}[{levelshortname!s}]{reset!s} {message!s}'

        class ColoredPrefixFormatter(colorlog.ColoredFormatter):
            def format(self, record: Any, *args: Any) -> Any:
                record.levelshortname = LEVEL_PREFIX[record.levelname]
                return super().format(record, *args)

        ch_fmt = ColoredPrefixFormatter(fmt, log_colors=LOG_COLORS, style='{')
    else:
        fmt = ' [{levelshortname!s}] {message!s}'

        class PrefixFormatter(logging.Formatter):
            def format(self, record: Any) -> Any:
                record.levelshortname = LEVEL_PREFIX[record.levelname]
                return super().format(record)

        ch_fmt = PrefixFormatter(fmt, style='{')

    ch.setFormatter(ch_fmt)
    package_logger.handlers = [ch]

    if catch_warnings:
        logging.captureWarnings(True)
        warnings_logger = logging.getLogger('py.warnings')
        warnings_logger.handlers = [ch]
        warnings_logger.setLevel(level)

    return package_logger



"""exceptions.py

Custom warnings and exceptions raised internally.
"""


class TileOutOfBoundsError(Exception):
    pass


class DatasetNotFoundError(Exception):
    pass


class InvalidKeyError(Exception):
    pass


class InvalidArgumentsError(Exception):
    pass


class InvalidDatabaseError(Exception):
    pass


class PerformanceWarning(UserWarning):
    pass
